# Day 0: VNF Instantiation & Management setup

## Description of this phase

The objective of this section is to provide the guidelines for including all the necessary elements in the VNF Package for its successful instantiation and management setup, so it can be further configured at later stages.

The way to achieve this in OSM is to prepare the descriptor so that it accurately details the VNF requirements, prepare cloud-init scripts (if needed), and identify parameters that may have to be provided at later stages to further adapt to different infrastructures.

## Day-0 Onboarding Guidelines

### Building the initial package

The most straightforward way to build a VNF package from scratch is to use the existing script available a the OSM Devops repository. From a Linux/Unix-based system:

#### Clone the OSM DevOps repository and access the tools folder.

```
git clone https://osm.etsi.org/gerrit/osm/devops.git
cd devops/descriptor-packages/tools
```

#### Run the generator script with the desired options.

`./generate_descriptor_pkg.sh [options] [name]`

Most common options are:

|     Parameter     |  Scope  |                 Description                 |   Values   |
|:-----------------:|:-------:|:-------------------------------------------:|:----------:|
|         -t        | package | descriptor type                             | vnfd | nsd |
|         -a        | package | create package for the descriptor           |      -     |
|         -N        | package | keep folder after tar is built              |      -     |
|         -c        | package | create folder structure inside package      |      -     |
|         -d        | package | destination of the folder                   |    path    |
|       --nsd       | package | create folder structure for NSD as well     |      -     |
|      --image      |   vdu   | image name                                  |    name    |
|       --vcpu      |   vdu   | vCPU number                                 |      #     |
|      --memory     |   vdu   | RAM size                                    |    [mb]    |
|     --storage     |   vdu   | disk size                                   |    [gb]    |
| --cloud-init-file |   vdu   | cloud-init file name                        |    name    |
|    --interfaces   |   vdu   | interface number (additional to management) |      #     |
|      --vendor     |   vnf   | vendor name                                 |    name    | 

For example:

```
./generate_descriptor_pkg.sh -t vnfd -N -c -d /home/ubuntu \
-a --image haproxy_ubuntu --vcpu 2 --memory 4096 --storage 10 \
--cloud-init-file init_lb --interfaces 2 --vendor ACME --nsd vLB
```

Note that we are adding the 'nsd' keyword to create also a NS Package that refers to this VNF Package, to be able to instantiate it and test it out. So the above example will create, in the `/home/ubuntu` folder:

- `vLB_vnfd` 				→ VNFD Folder
- `vLB_vnfd.tar.gz` 		→ VNFD Package
- `test_vnf01_nsd`			→ NSD Folder
- `test_vnf01_nsd.tar.gz`	→ NSD Package

**The VNFD Folder will contain the YAML file which models the VNF. This should be further edited to achieve the desired characteristics.**

### Modelling advanced topologies

Most topology types, along with sample descriptor files, can be found [here](https://osm.etsi.org/wikipub/index.php/Reference_VNF_and_NS_Descriptors).

When dealing with multiple VDUs inside a VNF, it is important to understand the differences between external and internal connection points (CPs) and virtual link descriptors (VLDs).

| Component | Definition | Modelled at |
|:------------:|---------------------------------------------------------------------------|:-----------:|
| Internal VLD | Network that interconnects VDUs within a VNF | VNFD |
| External VLD | Network that interconnects different VNFs within a NS | NSD |
| Internal CP | Element internal to a VNF, maps VDU interfaces to internal VLDs | VNFD |
| External CP | Element exposed externally by a VNF, maps VDU interfaces to external VLDs | NSD |

As VNF Package builders, we should clearly identify interfaces that i) are internal to the VNF and used to interconnect our own VDUs through internal VLDs, and ii) those we want to expose to other VNFs within a Network Service, using external VLDs.

In this example from the [5th OSM Hackfest](https://osm.etsi.org/wikipub/index.php/5th_OSM_Hackfest), we are building the following Multi-VDU topology:

![](assets/multi-vdu-sample.png)

The VNFD would look like this:

```
vnfd:vnfd-catalog:
    vnfd:
    -   ...
        # A external CP should be used for VNF management
        mgmt-interface:
            cp: vnf-mgmt

        # External CPs are exposed externally, to be referred at the NSD
        connection-point:
        -   id: vnf-mgmt
            name: vnf-mgmt
            short-name: vnf-mgmt
            type: VPORT
        -   id: vnf-data
            name: vnf-data
            short-name: vnf-data
            type: VPORT

        # Internal VLDs are defined globally at the VNFD
        internal-vld:
        -   id: internal
            name: internal
            short-name: internal
            type: ELAN
            internal-connection-point:
            -   id-ref: mgmtVM-internal
            -   id-ref: dataVM-internal

        # Inside the VDU block, multiple VDUs, their interfaces and CPs are modelled
        vdu:
        -   id: mgmtVM
            ...

            # VDU Interfaces map to either a external o internal CP
            interface:
            -   name: mgmtVM-eth0
                position: '1'
                type: EXTERNAL
                virtual-interface:
                    type: VIRTIO
                external-connection-point-ref: vnf-mgmt
            -   name: mgmtVM-eth1
                position: '2'
                type: INTERNAL
                virtual-interface:
                    type: VIRTIO
                internal-connection-point-ref: mgmtVM-internal

            # Internal CPs are modelled inside each VDU
            internal-connection-point:
            -   id: mgmtVM-internal
                name: mgmtVM-internal
                short-name: mgmtVM-internal
                type: VPORT

        -   id: dataVM
            ...
            # VDU Interfaces map to either a external o internal CP
            interface:
            -   name: dataVM-eth0
                position: '1'
                type: INTERNAL
                virtual-interface:
                    type: VIRTIO
                internal-connection-point-ref: dataVM-internal
            -   name: dataVM-xe0
                position: '2'
                type: EXTERNAL
                virtual-interface:
                    type: VIRTIO
                external-connection-point-ref: vnf-data

            # Internal CPs are modelled inside each VDU                
            internal-connection-point:
            -   id: dataVM-internal
                name: dataVM-internal
                short-name: dataVM-internal
                type: VPORT
```

As an additional reference, let's take a look at this Network Service Descriptor (NSD), where connections between VNFs are modelled using external CPs mapped to external VLDs like this:

```
nsd:nsd-catalog:
    nsd:
    -   ...
    	# External VLDs are modelled globally
        vld:
        -   id: mgmtnet
            name: mgmtnet
            short-name: mgmtnet
            type: ELAN
            mgmt-network: 'true'
            vim-network-name: mgmt
            vnfd-connection-point-ref:

            # Mapping between VNF's external CPs and the external VLD occurs here:
            -   vnfd-id-ref: hackfest_multivdu-vnf
                member-vnf-index-ref: '1'
                vnfd-connection-point-ref: vnf-mgmt
            -   vnfd-id-ref: hackfest_multivdu-vnf
                member-vnf-index-ref: '2'
                vnfd-connection-point-ref: vnf-mgmt
```

### Modelling specific networking requirements

Even though it is not recommended to hard-code networking values in order to maximize the VNF Package uniqueness, there may be some freedom for doing this at internal VLDs, especially when they are not externally accessible by other VNFs and not directly accessible from the management network.

The **IP Profiles** feature allows us to set some subnet specifics that can become useful. Further IP Profile settings can be found at the [OSM Information Model Documentation](https://osm.etsi.org/wikipub/index.php/OSM_Information_Model). The following VNFD extract can be used as a reference:

```
vnfd:vnfd-catalog:
    vnfd:
    -   ...
    	# IP profiles let us set subnet parameters like disabling a default GW
        ip-profiles:
        -   name: ip1
            description: ip1
            ip-profile-params:
                ip-version: ipv4
                dns-server: 8.8.8.8 
                gateway-address:     
                subnet-address: 192.168.100.0/24
                dhcp-params:
                  enabled: true

        # The IP Profile name is then applied at the VLD level
        internal-vld:
        -   id: internal
             ip-profile-ref: ip1
             ...
```

Specific IP and MAC addresses can also be set, although this practice is not recommended unless we use it in isolated connection points.

```
vnfd:vnfd-catalog:
    vnfd:
    -   ...
    	# A specific IP address can be set at the VLD, it requires the subnet to be predefined by using an IP Profile
        internal-vld:
        -   id: internal
            ip-profile-ref: p1
            ...
            internal-connection-point:
            -   id-ref: mgmtVM-internal
                ip-address: 192.168.100.100
        ...
        vdu:
        -   id: mgmtVM
            ...

            # A specific MAC address can also be set at the interface level
            interface:
            -   ...
                mac-address: '01:02:03:01:02:03'
```

### Building and adding cloud-init scripts

#### Cloud-init basics

Cloud-init is normally used for Day-0 operations like: 

* Setting a default locale
* Setting an instance hostname
* Generating instance SSH private keys or defining passwords
* Adding SSH keys to a user’s .ssh/authorized_keys so they can log in
* Setting up ephemeral mount points
* Configuring network devices
* Adding users and groups
* Adding files

Cloud-init scripts are referred at the VDU level. These can be defined inline or can be included in the **cloud_init** folder of the VNF package, then referred in the descriptor.

For inline cloud-init definition, follow this:

```
vnfd:vnfd-catalog:
  vnfd:
  - ...
    vdu:
    - ...
      cloud-init: |
        #cloud-config
        ...
```

For external cloud-init definition, proceed like this:

```
vnfd:vnfd-catalog:
  vnfd:
  - ...
    vdu:
    - ... 
      cloud-init-file: cloud_init_filename
```
Its content can have a number of formats, including #cloud-config and bash.
For example, any of the following scripts sets a password in Linux.

```
#cloud-config
hostname: lb_vdu
password: osm2018
chpasswd: { expire: False }
ssh_pwauth: True
```

```
#cloud-config
hostname: lb_vdu
chpasswd:
  list: |
    ubuntu:osm2018
  expire: False
```

Additional information about cloud-init can be found in [this documentation](http://cloudinit.readthedocs.io/en/latest/).

#### Parametrizing Cloud-init files

Beginning in OSM version 5.0.3, cloud-init files can be parametrized by using double curly brackets following Jinja2 sintaxis. For example:

```yaml
#cloud-config
hostname: lb_vdu
password: "{{ password }}"
chpasswd: { expire: False }
ssh_pwauth: True
```

Such variables can then be passed at instantiation time by referring the VNF index it applies to, as well as the name and value of the variable.

```
osm ns-create ... --config "{additionalParamsForVnf: [{member-vnf-index: '1', additionalParams:{password: 'secret'}}]}"
```

When dealing with multiple variables, it might be useful to pass a YAML file instead.

```
osm ns-create ... --config-file vars.yaml
```

Please note that variable naming convention follows Jinja2 (Python identifiers), so hyphens are not allowed.

In addition to the parameters provided by the user, OSM adds its own variable called 'OSM', that is an object with following content:

```yaml
OSM:
    ns_id:  OSM ns identity
    vnf_id: OSM vnf identity
    member_vnf_index:   map <NSD>.constituent-vnfd.member-vnf-index
    vdu_id: map <VNFD>.vdu.id
    count_index: vdu instance number, starting by 0
    ip_address: if known at instantiation time
    vdu:  # vdus where the key is the <vdu_id>-<count_index>
        id1-0:
            vdu_id: map <VNFD>.vdu.id
            count_index: vdu instance number, starting by 0
            ip_address: if known at instantiation time, that is, if forced
            interfaces:  # vdu interfaces where the key is the interface name
                name1:
                    ip_address: if known at instantiation time
                    mac_address: if known at instantiation time
                    vnf_vld_id: id of the <VNFD>.vld.id where this interface is connected
                    name: map <VNFD>.vdu.interfaces.name
```

Apart from 

#### Support for Configuration Drive

Besides cloud-init being provided as userdata through a metadata service, some VNFs will require to store the metadata locally on a configuration drive.

The support for this is available at the VNFD model, as follows:

```
supplemental-boot-data:
    boot-data-drive: 'true'
```

### Guidelines for EPA requirements

Most EPA features can be specified at the VDU descriptor level as requirements in the `guest-epa` section, which will be then translated to the appropriate request through the VIM connector. Please note that the NFVI should be pre-configured to support these EPA capabilities.

#### Huge Pages

Huge pages are requested as follows:

```
vnfd:vnfd-catalog:
  vnfd:
  - ...
    vdu:
    -   ...
        guest-epa:
            mempage-size: LARGE
            ...
```

The `mempage-size` attribute can take any of these values:

* LARGE: Require hugepages (either 2MB or 1GB)
* SMALL: Doesn't require hugepages
* SIZE_2MB: Requires 2MB hugepages
* SIZE_1GB: Requires 1GB hugepages
* PREFER_LARGE: Application prefers hugepages

#### CPU Pinning

CPU pinning allows for different settings related to vCPU assignment and hyper threading:

```
vnfd:vnfd-catalog:
  vnfd:
  - ...
    vdu:
    -   ...
        guest-epa:
            cpu-pinning-policy: DEDICATED
            cpu-thread-pinning-policy: AVOID
            ... 
```

CPU pinning policy describes association between virtual CPUs in guest and the
physical CPUs in the host. Valid values are:

* DEDICATED: Virtual CPUs are pinned to physical CPUs
* SHARED: Multiple VMs may share the same physical CPUs.
* ANY: Any policy is acceptable for the VM

CPU thread pinning policy describes how to place the guest CPUs when the host supports hyper threads. Valid values are:

* AVOID: Avoids placing a guest on a host with threads.
* SEPARATE: Places vCPUs on separate cores, and avoids placing two vCPUs on two threads of same core.
* ISOLATE: Places each vCPU on a different core, and places no vCPUs from a different guest on the same core.
* PREFER: Attempts to place vCPUs on threads of the same core.

#### NUMA Topology Awareness

This policy defines if the guest should be run on a host with one NUMA node or multiple NUMA nodes.

```
vnfd:vnfd-catalog:
  vnfd:
  - ...
    vdu:
    -   ...
        guest-epa:
            numa-node-policy:
                node-cnt: 2
                mem-policy: STRICT
                node:
                    - id: 0
                      memory-mb: 2048
                      num-cores: 1
                    - id: 1
                      memory-mb: 2048
                      num-cores: 1                   
            ... 
```

`node-cnt` defines the number of NUMA nodes to expose to the VM, while `mem-policy` defines if the memory should be allocated strictly from the 'local' NUMA node (STRICT) or not necessarily from that node (PREFERRED).
The rest of the settings request a specific mapping between the NUMA nodes and the VM resources and can be explored at detail in the [OSM Information Model Documentation](https://osm.etsi.org/wikipub/index.php/OSM_Information_Model)

#### SR-IOV and PCI-Passthrough

Dedicated interface resources can be requested at the VDU interface level.

```
vnfd:vnfd-catalog:
  vnfd:
  - ...
    vdu:
    -   ...
        interface:
        -   name: eth0
            position: '1'
            type: EXTERNAL
            virtual-interface:
                type: SR-IOV            
```

Valid values for `type`, which specifies the type of virtual interface between VM and host, are:

* PARAVIRT        : Use the default paravirtualized interface for the VIM (virtio, vmxnet3, etc.).
* PCI-PASSTHROUGH : Use PCI-PASSTHROUGH interface.
* SR-IOV          : Use SR-IOV interface.
* E1000           : Emulate E1000 interface.
* RTL8139         : Emulate RTL8139 interface.
* PCNET           : Emulate PCNET interface.

### Managing alternative images for specific VIM types

The `image` name specified at the VDU level is expected to be either located at the `images` folder within the VNF package, or at the VIM catalogue.

Alternative images can be specified and mapped to different VIM types, so that they are used whenever the VNF package is instantiated over the given VIM type.
In the following example, the `ubuntu1604` image is used by default (any VIM), but a different image is used if the VIM type is AWS.

```
vnfd:vnfd-catalog:
  vnfd:
  - ...
    vdu:
    -   ...
        image: ubuntu1604
        alternative-images:
        -   vim-type: aws
            image: ubuntu/images/hvm-ssd/ubuntu-artful-17.10-amd64-server-20180509
```

### Updating and Testing Instantiation of the VNF Package

Once the VNF Descriptor has been updated with all the Day-0 requirements, its folder needs to be repackaged. For example, in Linux/UNIX, it would be something like: `tar -cvfz vLB_vnfd.tar.gz vLB_vnfd/`

A Network Service package containing at least this single VNF needs to be used to instantiate the VNF. This could be generated with the _devops tool_ described earlier. 

Remember the objectives of this phase:
1. Instantiating the VNF with all the required VDUs, images, initial (unconfigured) state and NFVI requirements.
2. Making the VNF manageable from OSM (OSM should have SSH access to the management interfaces, for example)

To test this out, the NS can be launched using the OSM client, like this:
```
osm ns-create --ns_name [ns name] --nsd_name [nsd name] --vim_account [vim name] \
--ssh_keys [comma separated list of public key files to inject to vnfs]
```

At launch time, extra **_instantiation parameters_** can be passed so that the VNF can be adapted to the particular instantiation environment or to achieve a proper inter-operation with other VNFs within the specific NS.
More information about these parameters will be revised during the next chapter as part of Day-1 objectives, or can be reviewed [here](https://osm.etsi.org/wikipub/index.php/OSM_instantiation_parameters).

The following sections will provide details on how to further populate the VNF Package to automate Day 1/2 operations.
