# Day 2: VNF Runtime Operations

## Description of this phase

The objective of this section is to provide the guidelines for including all the necessary elements in the VNF Package so that it can be operated at runtime and therefore, reconfigured on demand at any point by the end-user. Typical operations include reconfiguration of services, KPI monitoring and the enablement of automatic, closed-loop operations triggered by monitored status.

The main mechanisms to achieve reconfiguration in OSM is to build a Proxy Charm and include it in the descriptor. On the other hand, monitoring and VNF-specific policy management can be achieved by specifying the requirements at the descriptor (modifying monitored indicators and policies at runtime is not supported in OSM as of version 5.0.5)

## Day-2 Onboarding Guidelines

### Adding Day-2 primitives to the descriptor

Day-2 primitives are actions invoked on demand, so the `config-primitive` block is used instead of the `initial-config-primitive` block at the VNF or VDU level.

For example, a VNF-level set of Day-2 primitives would look like this:

```
vnfd:vnfd-catalog:
  vnfd:
  - ...
    mgmt-interface:
        cp: vnf-cp0
    ...
    vnf-configuration:
        ...
        config-primitive:
        -   name: restart-service
            parameter:
            -   name: offset
                default-value: 10
                data-type: STRING
        -   name: clean-cache
            parameter:
            -   name: force
                default-value: true
                data-type: BOOLEAN              
        juju:
            charm: samplecharm
```

### Building a Proxy Charm

Proxy charms for implementing Day-2 primitives are built exactly in the same way as when implementing Day-1 primitives.

### Adding monitoring parameters

#### Collecting NFVI metrics

In order to collect NFVI-level metrics associated to any given VDU and store them in the OSM TSDB (using Prometheus software), a set of `monitoring-params` should be declared both globally and at the VDU level. 

Only CPU and Memory are supported as of OSM version 5.0.5. For example:

```
vnfd:vnfd-catalog:
  vnfd:
  - ...
    vdu:
    - id: "apache_vdu"
      ...  
      monitoring-param:
        - id: "apache_cpu_util"
          # nfvi-metric name should match the supported set of metrics collectable by OSM through the VIM connectors
          nfvi-metric: "cpu_utilization"
        - id: "apache_memory_util"
          nfvi-metric: "average_memory_utilization"          
    ...
    monitoring-param:
    -   id: "apache_vnf_cpu_util"
        name: "apache_vnf_cpu_util"
        # only 'AVERAGE' aggregation is supported at this time
        aggregation-type: AVERAGE
        vdu-monitoring-param:
          # vdu-ref should match the id of the VDU
          vdu-ref: "apache_vdu"
          # vdu-monitoring-param-ref should match the nfvi-metric id
          vdu-monitoring-param-ref: "apache_cpu_util"
    -   id: "apache_vnf_memory_util"
        name: "apache_vnf_memory_util"
        aggregation-type: AVERAGE
        vdu-monitoring-param:
          vdu-ref: "apache_vdu"
          vdu-monitoring-param-ref: "apache_memory_util"           
```

#### Collecting VNF indicators

As of OSM version 5.0.5, collection of VNF indicators is done by using Proxy Charms with the *metrics layer*. This is a simple method that has a couple of limitations:

* Metrics are collected every five minutes and this can't be changed.
* Only positive decimal values of *gauge* or *absolute* types can be collected.

At the charm level, the only file that needs to be created before building it is the "metrics.yaml" file at the root folder of the charm.

For example, the following file collects *active users* and *loads* values from a Linux machine.

```
# metrics.yaml file
metrics:
   users:
     type: gauge
     description: "# of users"
     command: who|wc -l
   load:
     type: gauge
     description: "5 minute load average"
     command: cat /proc/loadavg |awk '{print $1}'         
```

More information on how to populate this file can be found in the Juju [developer metrics](https://docs.jujucharms.com/2.5/en/developer-metrics) documentation.

Once the charm has been created and included in the VNF Package, the descriptor needs to define the metrics to be actually collected by OSM. As with any charm, this can be done at a VNF or VDU level.

For example, at the VNF level (a VDU that represents the VNF):

```
vnfd:vnfd-catalog:
  vnfd:
  - ...
    mgmt-interface:
        cp: vnf-cp0
    ...
    vnf-configuration:
      ...
      juju:
        # this is the name of the proxy charm
        charm: metricscharm
      metrics:
        # metric names should match the ones specified at the metrics.yaml file
        - name: users
        - name: load
    ...
    monitoring-param:
    -   id: "ubuntuvdu_users"
        name: "ubuntuvdu_users"
        aggregation-type: AVERAGE
        vnf-metric:
          # vnf-metric-name-ref should match the metric name specified at VNF/VDU level
          vnf-metric-name-ref: "users"
    -   id: "ubuntuvdu_load"
        name: "ubuntuvdu_load"
        aggregation-type: AVERAGE
        vnf-metric:
          vnf-metric-name-ref: "load"
```

This other example does the same, but at a specific VDU level:

```
vnfd:vnfd-catalog:
  vnfd:
  - ...
    vdu:
    - ...
      interface:
      # remember that for any VDU that has runs a charm, a management interface needs to be specified
      - external-connection-point-ref: vdu1_mgmt
        mgmt-interface: true
      ...
      vdu-configuration:
        ...
        juju:
          charm: metricscharm
        metrics:
          - name: users
          - name: load
    ...   
    monitoring-param:
    -   id: "ubuntuvdu_users"
        name: "ubuntuvdu_users"
        aggregation-type: AVERAGE
        vdu-metric:
          vdu-ref: "ubuntuvdu1"
          vdu-metric-name-ref: "users"
    -   id: "ubuntuvdu_load"
        name: "ubuntuvdu_load"
        aggregation-type: AVERAGE
        vdu-metric:
          vdu-ref: "ubuntuvdu1"
          vdu-metric-name-ref: "load"
```

### Adding scaling operations

Scaling operations happen at a VDU level and can be added with automatic triggers (*closed-loop* mode triggered by *monitoring-param* thresholds), or with a manual trigger.

In both cases, a `scaling-group-descriptor` section must be added to the VNF descriptor. The following example enables VDU scaling based on a manual trigger (OSM API or CLI).

```
vnfd:vnfd-catalog:
  vnfd:
  - ...                     
    scaling-group-descriptor:
    -   name: "apache_vdu_manualscale"
        # the following counts refer to "scaled instances" only
        min-instance-count: 0
        max-instance-count: 10
        scaling-policy:
        -   name: "manual_policy"
            scaling-type: "manual"
        vdu:
        -   vdu-id-ref: apache_vdu
            count: 1
```

The following example defines a closed-loop scaling operation based on a specific monitoring parameter threshold.

```
vnfd:vnfd-catalog:
  vnfd:
  - ...                     
    scaling-group-descriptor:
    -   name: "apache_vdu_autoscale"
        min-instance-count: 0
        max-instance-count: 10
        scaling-policy:
        -   name: "apache_cpu_util_above_threshold"
            scaling-type: "automatic"
            threshold-time: 10
            cooldown-time: 120
            scaling-criteria:
            -   name: "apache_cpu_util_above_threshold"
                # this is the name of the monitoring-param to monitor
                vnf-monitoring-param-ref: "apache_vnf_cpu_util"                
                # scale-in threshold
                scale-in-threshold: 20
                scale-in-relational-operation: "LT"
                # scale-out threshold
                scale-out-threshold: 80
                scale-out-relational-operation: "GT"
        vdu:
        -   vdu-id-ref: apache_vdu
            count: 1
```

More information about scaling can be found in the [OSM Autoscaling documentation](https://osm.etsi.org/wikipub/index.php/OSM_Autoscaling)

### Testing Instantiation of the VNF Package

Each of the objectives of this phase can be tested as follows:

* **Enabling a way of re-configuring the VNF on demand**: primitives can be called through the OSM API, dashboard, or directly by running the following OSM client command: `osm ns-action [ns-name] --vnf_name [vnf-index] --action_name [primitive-name] --params '{param-name-1: "param-value-1", param-name-2: "param-value-2", ...}`

* **Monitor the main KPIs of the VNF**: if correctly enabled, metrics will automatically start appearing in the OSM Prometheus database. More information on how to access, visualize and troubleshoot metrics can be found in the [OSM Performance Management documentation](https://osm.etsi.org/wikipub/index.php/OSM_Performance_Management)

* **Enabling scaling operations**: automatic scaling should be tested by making the metric reach the corresponding threshold, while manual scaling can be tested by using the following command (which also works when the "scaling-type" has been set to "automatic"): `osm vnf-scale [ns-name] [vnf-name] --scaling-group [scaling-group name] [--scale-in|--scale-out]`
